import functools
import inspect
from collections import deque
from copy import copy
from typing import Mapping

from pandaSuit.common.mappings.reversible import *
from pandaSuit.common.unwind import Unwind
from pandaSuit.common.constant.decorators import UNWIND_LIST


def extract_function_name(function_reference: str) -> str:
    return function_reference.split(".")[1].split(" at")[0]


def in_place_operation(args: tuple, kwargs: dict, signature: inspect.Signature) -> bool:
    # search for in_place parameter in positional arguments
    if len(args) > 0:
        for count, arg in enumerate(args):
            if list(signature.parameters.keys())[count] == "in_place":  # if in_place parameter is passed
                return arg

    # search for in_place parameter in keyword arguments
    if kwargs.get("in_place") is not None:
        return kwargs.get("in_place")

    # otherwise, use default for in_place parameter
    return signature.parameters.get("in_place").default


def infer_kwargs(args: tuple, signature: inspect.Signature) -> dict:
    kwargs = {}
    for count, arg in enumerate(args):
        kwargs[list(signature.parameters.keys())[count]] = args[count]
    return kwargs


def get_method_signature(df_object: object, name: str) -> inspect.Signature:
    return inspect.signature(df_object.__getattribute__(name))


def get_method_parameters(df_object: object, name: str) -> Mapping[str, inspect.Parameter]:
    return get_method_signature(df_object, name).parameters


def reversible(func):
    """Allow for reversing an 'in place' operation on pandaSuit object"""
    @functools.wraps(func)
    def wrapper_reverse(*args, **kwargs):
        caller_function = inspect.stack()[1][3]
        if caller_function != "undo":  # this occurs when a @reversible method is un-done by another @reversible method
            function_name = extract_function_name(func.__repr__())
            df_object = args[0]
            method_signature = get_method_signature(df_object, function_name)
            if not in_place_operation(args[1:], kwargs, method_signature):
                return func(*args, **kwargs)  # don't create unwind step, but return value from method called
            else:
                if len(args) > 1:  # convert positional args into keyword args
                    kwargs.update(infer_kwargs(args[1:], method_signature))
                    args = (df_object,)  # remove positional args to avoid passing parameters multiple times when calling the function
                intermediate_reverse_function = INTERMEDIATE_REVERSE_FUNCTION_MAPPING.get(function_name)
                reverse_args = REVERSE_ARGS.get(function_name)(df=df_object, arguments=kwargs)
                if intermediate_reverse_function is not None:  # some DF manipulations require an intermediate step to reverse (e.g., .update() & .remove())
                    intermediate_reverse_args = INTERMEDIATE_REVERSE_ARGS.get(function_name)(kwargs)
                    reverse_args_to_add = df_object.__getattribute__(intermediate_reverse_function)(**intermediate_reverse_args)
                    intermediate_arg_mapping_function = INTERMEDIATE_ARGUMENT_MAPPING.get(function_name)
                    reverse_args.update(intermediate_arg_mapping_function(copy(kwargs), copy(reverse_args_to_add)))
                reverse_function = REVERSE_MAPPING.get(function_name)
                if "in_place" in get_method_parameters(df_object, reverse_function):
                    reverse_args.update({"in_place": True})
                df_object.__setattr__(UNWIND_LIST, df_object.__getattribute__(UNWIND_LIST) + deque([Unwind(reverse_function, reverse_args)]))
        func(*args, **kwargs)
    return wrapper_reverse



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from pandas import Series

from pandaSuit.common.constant.plot import MULTI_BAR_PLOT_TYPES
from pandaSuit.plot.plot import Plot


class BarPlot(Plot):
    def __init__(self,
                 x: Series or dict or list,
                 y: Series or dict or list,
                 multi_bar_type: str = "beside",
                 bar_width: float or int = None,
                 **kwargs):
        self._shown = False
        self.x = x
        self.y = y
        if isinstance(self.y, list) and len(self.y) > 1:
            if multi_bar_type is not None:
                if multi_bar_type in MULTI_BAR_PLOT_TYPES:
                    self.multi_bar_type = multi_bar_type
                else:
                    raise ValueError("Must specify an alternate type of bar plot when supplying multiple "
                                     "y arguments (e.g. stacked, beside, etc.)")
            else:
                raise ValueError("Must specify an alternate type of bar plot when supplying multiple "
                                 "y arguments (e.g. stacked, beside, etc.)")
        else:
            self.multi_bar_type = None
        self.bar_width = bar_width if bar_width is not None else 0.9/len(y)
        self.figure, self.axes = plt.subplots()
        self.title = kwargs.get('title')
        self.x_label = kwargs.get('x_label')
        self.y_label = kwargs.get('y_label')
        self.legend = kwargs.get('legend')
        super().__init__(self.figure, self.axes)

    def display(self) -> None:
        if self._shown:
            self = BarPlot(x=self.x,
                           y=self.y,
                           multi_bar_type=self.multi_bar_type,
                           bar_width=self.bar_width,
                           title=self.title,
                           x_label=self.x_label,
                           y_label=self.y_label,
                           legend=self.legend)
            self.display()
        else:
            self.create_plot()
            self._shown = True
            plt.show()

    def create_plot(self) -> None:
        if self.multi_bar_type is None:
            self._construct_base_plot()
        elif self.multi_bar_type == "beside":
            self._construct_side_by_side_plot()
        elif self.multi_bar_type == "stacked":
            self._construct_stacked_plot()
        else:
            pass  # todo: add error handler
        self._add_plot_features()

    def _add_plot_features(self) -> None:
        if self.title is not None:
            self.axes.set_title(self.title)
        if self.x_label is not None:
            self.axes.set_xlabel(self.x_label)
        if self.y_label is not None:
            self.axes.set_ylabel(self.y_label)
        if self.legend:
            self.axes.legend()

    def _construct_base_plot(self) -> None:
        self.axes.bar(self.x if isinstance(self.x, list) else self.x.to_list(), self.y[0].to_list(),
                      label=self.y[0].name)

    def _construct_side_by_side_plot(self) -> None:
        x_indexes = np.arange(len(self.x))
        counter = 0
        for y in self.y:
            self.axes.bar(x_indexes-(self.bar_width*counter),
                          y if isinstance(y, list) else y.to_list(),
                          width=self.bar_width,
                          label=y.name)
            counter += 1
        plt.xticks(ticks=x_indexes, labels=self.x if isinstance(self.x, list) else self.x.to_list())

    def _construct_stacked_plot(self) -> None:
        for y in self.y:
            self.axes.bar(self.x if isinstance(self.x, list) else self.x.to_list(),
                          y if isinstance(y, list) else y.to_list(),
                          label=y.name)



import numpy as np


def find_index(list_to_search: list, value: object) -> list:
    return find_indexes(list_to_search, value)[0]


def find_indexes(list_to_search: list, value: object) -> list:
    return [index for index, element in enumerate(list_to_search) if element == value]


def index_dictionary(list_of_values: list) -> dict:
    return {value: find_indexes(list_of_values, value) for value in list_of_values}


def create_index_list(stop: int, start: int = 0, step: int = 1) -> list:
    return list(map(int, np.arange(start=start, stop=stop, step=step)))


def remove_elements(base_list: list, exclude: list or object) -> list:
    return [element for element in base_list if element not in exclude] \
        if isinstance(exclude, list) else [element for element in base_list if element != exclude]


def remove_elements_from_dict(base_dict: dict, exclude: list or object) -> dict:
    return {key: value for key, value in base_dict.items() if key not in exclude} \
        if isinstance(exclude, list) else {key for key, value in base_dict.items() if key != exclude}


def combinations(list1, list2):
    return [(i, j) for j in list2 for i in list1]


def filter_and_find_indexes(base_list: list, exclude: list or object) -> list:
    return [find_indexes(base_list, element)[0] for element in base_list if element not in exclude] \
        if isinstance(exclude, list) else \
        [find_indexes(base_list, element)[0] for element in base_list if element != exclude]


def is_continuous_list(int_list: list) -> bool:
    last_int, differences_list = int_list[0], []
    for integer in int_list[1:]:
        differences_list.append(integer - last_int)
        last_int = integer
    return all(value == 1 for value in differences_list)



from sklearn.linear_model import LinearRegression
from pandas import Series, DataFrame


class LinearModel:

    def __init__(self, dependent: Series or DataFrame, independent: Series or DataFrame, intercept: bool = True):
        super().__init__()
        self.dependent = dependent
        self.independent = independent
        self.include_intercept = intercept
        self.model = self._fit()

    def predict(self, input: Series or dict or int or float, practical: bool = False) -> float or list:
        return self._practical_prediction(input) if practical else self._theoretical_prediction(input)

    def _fit(self) -> LinearRegression:
        if isinstance(self.dependent, DataFrame):
            y = self.dependent.values
        else:
            y = self.dependent.to_list()
        if isinstance(self.independent, Series):
            x = self.independent.to_numpy().reshape(-1, 1)
        else:
            x = self.independent
        return LinearRegression(fit_intercept=self.include_intercept).fit(X=x, y=y)

    # Properties
    @property
    def intercept(self) -> float:
        return self.model.intercept_

    @property
    def betas(self) -> list:
        return self.model.coef_.tolist()[0]

    # Private prediction methods
    def _theoretical_prediction(self, input: Series or dict) -> float:
        if isinstance(input, Series):
            return list(self.model.predict(input.to_numpy().reshape(1, -1))[0])
        else:
            return list(self.model.predict(Series(input).to_numpy().reshape(1, -1))[0])

    def _practical_prediction(self, input: Series or dict) -> float:  # todo
        if isinstance(input, dict):
            return (self.betas * Series(input)).sum() + (self.intercept if self.include_intercept else 0.0)
        elif isinstance(input, Series):
            return (self.betas * input).sum() + (self.intercept if self.include_intercept else 0.0)
        elif isinstance(input, (float, int)):
            return (sum(self.betas * input)) + (self.intercept if self.include_intercept else 0.0)
        else:
            raise Exception("Must supply Series, dict, int or float to .predict() method")



from __future__ import annotations

import matplotlib.pyplot as plt
from pandas import Series

from pandaSuit.plot.plot import Plot


class LinePlot(Plot):
    def __init__(self, x: Series or dict or list, y: Series or dict or list, **kwargs):
        self._shown = False
        self.x = x if isinstance(x, Series) else Series(x)
        self.y = y if isinstance(y, Series) else Series(y)
        self.figure, self.axes = plt.subplots()
        self.title = kwargs.get('title')
        self.x_label = kwargs.get('x_label')
        self.y_label = kwargs.get('y_label')
        self.x_ticks = kwargs.get('x_ticks')
        self.x_bottom = kwargs.get("x_bottom")
        self.x_top = kwargs.get("x_top")
        self.y_bottom = kwargs.get("y_bottom")
        self.y_top = kwargs.get("y_top")
        self.legend = kwargs.get('legend')
        self.grid_lines = kwargs.get('grid_lines')
        self.x_scale = kwargs.get('x_scale')
        self.y_scale = kwargs.get('y_scale')
        super().__init__(self.figure, self. axes)

    def display(self) -> None:
        if self._shown:
            self = LinePlot(self.x, self.y,
                            title=self.title,
                            x_label=self.x_label,
                            y_label=self.y_label,
                            x_ticks=self.x_ticks,
                            x_bottom=self.x_bottom,
                            x_top=self.x_top,
                            y_bottom=self.y_bottom,
                            y_top=self.y_top,
                            legend=self.legend,
                            grid_lines=self.grid_lines,
                            x_scale=self.x_scale,
                            y_scale=self.y_scale)
            self.display()
        else:
            self.create_plot()
            self._shown = True
            if self.x_scale:
                plt.xscale(self.x_scale)
            if self.y_scale:
                plt.yscale(self.y_scale)
            plt.show()

    def create_plot(self) -> None:
        for y in self.y:
            self.axes.plot(self.x.to_list(), y.to_list(), label=y.name)
        self._add_chart_features()

    def _add_chart_features(self) -> None:
        if self.title is not None:
            self.axes.set_title(self.title)
        if self.x_label is not None:
            self.axes.set_xlabel(self.x_label)
        if self.y_label is not None:
            self.axes.set_ylabel(self.y_label)
        if self.x_ticks is not None:
            self.axes.set_xticklabels(self.x_ticks)
        if self.x_bottom is not None:
            self.axes.set_xlim(bottom=self.x_bottom)
        if self.x_top is not None:
            self.axes.set_xlim(top=self.x_top)
        if self.y_bottom is not None:
            self.axes.set_ylim(bottom=self.y_bottom)
        if self.y_top is not None:
            self.axes.set_ylim(top=self.y_top)
        if self.legend:
            self.axes.legend()
        self.axes.grid(False if self.grid_lines is None else self.grid_lines)



from __future__ import annotations

import matplotlib.pyplot as plt
from pandas import Series

from pandaSuit.plot.plot import Plot
from pandaSuit.stats.linear import LinearModel


class ScatterPlot(Plot):
    def __init__(self, x: Series or dict or list, y: Series or dict or list, **kwargs):
        self._shown = False
        self.x = x
        self.y = y
        self.figure, self.axes = plt.subplots()
        self.title = kwargs.get('title')
        self.x_label = kwargs.get('x_label')
        self.y_label = kwargs.get('y_label')
        self.best_fit_line = kwargs.get('best_fit_line')
        self.legend = kwargs.get('legend')
        super().__init__(self.figure, self.axes)

    def display(self) -> None:
        if self._shown:
            self = ScatterPlot(self.x, self.y,
                               title=self.title,
                               x_label=self.x_label,
                               y_label=self.y_label,
                               best_fit_line=self.best_fit_line,
                               legend=self.legend)
            self.display()
        else:
            self.create_plot()
            self._shown = True
            plt.show()

    def create_plot(self) -> None:
        for y in self.y:
            self.axes.scatter(self.x if isinstance(self.x, list) else self.x.to_list(),
                              y if isinstance(y, list) else y.to_list(), label=y.name)
        self._add_chart_features()

    def _add_chart_features(self) -> None:
        if self.title is not None:
            self.axes.set_title(self.title)
        if self.x_label is not None:
            self.axes.set_xlabel(self.x_label)
        if self.y_label is not None:
            self.axes.set_ylabel(self.y_label)
        if self.best_fit_line:
            for y in self.y:
                self.axes.plot(self._create_best_fit_line(y))
        if self.legend:
            self.axes.legend()

    def _create_best_fit_line(self, y_values: Series) -> list:
        x_values = self._prepare_x_axis_for_regression()
        lm = LinearModel(y_values, x_values)
        return [lm.predict(x) for x in x_values.to_list()]

    def _prepare_x_axis_for_regression(self) -> Series:
        return Series(name=self.x.name, data=[i for i, _ in enumerate(self.x)]) \
            if any(isinstance(x, str) for x in self.x) else self.x



from copy import copy

import pandas

from pandaSuit.common.constant.decorators import UPDATE, APPEND, INSERT, REMOVE, SELECT, RESET
from pandaSuit.common.util.list_operations import create_index_list, find_index, find_indexes


def intermediate_update_args(kwargs: dict) -> dict:
    return {
        "column": kwargs.get("column"),
        "row": kwargs.get("row"),
        "pandas_return_type": True
    }


def intermediate_remove_args(kwargs: dict) -> dict:
    return {
        "column": kwargs.get("column"),
        "row": kwargs.get("row"),
        "pandas_return_type": True
    }


def update_args(**kwargs) -> dict:
    return {
        "column": kwargs.get("arguments").get("column"),
        "row": kwargs.get("arguments").get("row"),
        "to": kwargs.get("arguments").get("to")
    }


def append_args(**kwargs) -> dict:
    if "row" in kwargs.get("arguments"):
        return {"row": -1}
    else:
        return {"column": -1}


def insert_args(**kwargs) -> dict:
    index = kwargs.get("arguments").get("index")
    row_arg, column_arg = kwargs.get("arguments").get("row"), kwargs.get("arguments").get("column")
    if row_arg is not None:
        if isinstance(row_arg, pandas.DataFrame) and isinstance(index, int):
            return {"row": create_index_list(start=index, stop=index+row_arg.shape[0])}
        else:
            return {"row": index}
    else:
        if isinstance(column_arg, pandas.DataFrame) and isinstance(index, int):
            return {"column": create_index_list(start=index, stop=index+column_arg.shape[1])}
        else:
            return {"column": index}


def remove_args(**kwargs) -> dict:
    row_arg = kwargs.get("arguments").get("row")
    column_arg = kwargs.get("arguments").get("column")
    if row_arg is not None:
        if isinstance(row_arg, str):
            return {"index": find_index(kwargs.get("df").__getattribute__("row_names"), row_arg)}
        else:  # list
            if isinstance(row_arg, list) and any(isinstance(row, str) for row in row_arg):  # row names
                return {"index": [find_index(kwargs.get("df").__getattribute__("row_names"), row) for row in row_arg]}
            else:  # row indexes
                return {"index": row_arg}
    elif column_arg is not None:
        if isinstance(column_arg, str):
            return {"index": find_index(kwargs.get("df").__getattribute__("column_names"), column_arg)}
        else:  # list
            if isinstance(column_arg, list) and any(isinstance(column, str) for column in column_arg):  # column names
                return {"index": [find_index(kwargs.get("df").__getattribute__("column_names"), column) for column in column_arg]}
            else:  # column indexes
                return {"index": column_arg}


def reset_args(**kwargs) -> dict:
    return {"data": copy(kwargs.get("df"))}


REVERSE_MAPPING = {
    UPDATE: UPDATE,
    APPEND: REMOVE,
    INSERT: REMOVE,
    REMOVE: INSERT,
    RESET: "_set_underlying_dataframe"
}

REVERSE_ARGS = {
    UPDATE: update_args,
    APPEND: append_args,
    INSERT: insert_args,
    REMOVE: remove_args,
    RESET: reset_args
}

INTERMEDIATE_REVERSE_FUNCTION_MAPPING = {
    UPDATE: SELECT,
    REMOVE: SELECT
}

INTERMEDIATE_REVERSE_ARGS = {
    UPDATE: intermediate_update_args,
    REMOVE: intermediate_remove_args
}


def update_reverse_args(kwargs: dict, old_value: object) -> dict:
    kwargs.update({"to": old_value})
    return kwargs


def remove_reverse_args(kwargs: dict, old_value: object) -> dict:
    if "row" in kwargs:
        kwargs.update({"row": old_value})
    else:
        kwargs.update({"column": old_value})
    return kwargs


INTERMEDIATE_ARGUMENT_MAPPING = {
    UPDATE: update_reverse_args,
    REMOVE: remove_reverse_args
}



from matplotlib import pyplot as plt
from pandas import Series

from pandaSuit.plot.plot import Plot


class Histogram(Plot):
    def __init__(self, y: Series or dict or list, **kwargs):
        self._shown = False
        self.y = y
        self.figure, self.axes = plt.subplots()
        self.title = kwargs.get('title')
        self.x_label = kwargs.get('x_label')
        self.y_label = kwargs.get('y_label')
        self.legend = kwargs.get('legend')
        self.bins = kwargs.get('bins')
        super().__init__(self.figure, self.axes)

    def display(self) -> None:
        if self._shown:
            self = Histogram(self.y,
                             title=self.title,
                             x_label=self.x_label,
                             y_label=self.y_label,
                             legend=self.legend)
            self.display()
        else:
            self.create_plot()
            self._shown = True
            plt.show()

    def create_plot(self) -> None:
        self.axes.hist(self.y, bins=self.bins)
        self._add_chart_features()

    def _add_chart_features(self) -> None:
        if self.title is not None:
            self.axes.set_title(self.title)
        if self.x_label is not None:
            self.axes.set_xlabel(self.x_label)
        if self.y_label is not None:
            self.axes.set_ylabel(self.y_label)
        if self.legend:
            self.axes.legend()



ALPHABET = "abcdefghijklmnopqrstuvwxyz"
DISTRIBUTIONS = [
    "exponential",
    "gumbel",
    "laplace",
    "logistic",
    "lognormal",
    "normal",
    "poisson",
    "rayleigh",
    "standard_cauchy",
    "standard_exponential",
    "standard_normal",
    "uniform"
]



from __future__ import annotations

from collections import deque
from copy import copy

import pandas as pd
from numpy import random as np_random

from pandaSuit.common.constant.date_constants import DATE_GROUPINGS
from pandaSuit.common.constant.df import ALPHABET, DISTRIBUTIONS
from pandaSuit.common.decorators import reversible
from pandaSuit.common.unwind import Unwind
from pandaSuit.common.util.list_operations import index_dictionary, create_index_list, find_indexes, is_continuous_list
from pandaSuit.plot.bar import BarPlot
from pandaSuit.plot.histogram import Histogram
from pandaSuit.plot.line import LinePlot
from pandaSuit.plot.pie import PiePlot
from pandaSuit.plot.scatter import ScatterPlot
from pandaSuit.stats.linear import LinearModel
from pandaSuit.stats.logistic import LogisticModel


class DF(pd.DataFrame):
    def __init__(self, data=None, overwrite_data=None):
        if overwrite_data is None:
            super().__init__(data=data)
            self.data = data
            self._unwind = deque()
        else:
            super().__init__(data=overwrite_data)  # this allows for self.__init__ calls while maintaining initial state and unwind steps

    def select(self,
               row: list or int or str = None,
               column: list or int or str = None,
               pandas_return_type: bool = False) -> pd.DataFrame or pd.Series or DF:
        if row is None:
            if self._names_supplied(column):
                result = self[column]
            else:
                result = self.iloc[:, column]
        else:
            if column is None:
                if self._names_supplied(row):
                    result = self.loc[row]
                else:
                    result = self.iloc[row]
            else:
                if self._names_supplied(row) and self._names_supplied(column):
                    result = self.loc[row, column]
                else:
                    if self._names_supplied(row):
                        result = self.loc[row, :][column]
                    else:
                        if self._names_supplied(column):
                            result = self.iloc[row, :][column]
                        else:
                            result = self.iloc[row, column]

        return self._df_query_return(result, pandas_return_type)

    def slice(self,
              from_row: int or str = 0,
              to_row: int or str = -1,
              from_column: int or str = 0,
              to_column: int or str = -1,
              pandas_return_type: bool = False) -> pd.DataFrame or pd.Series or DF:

        if isinstance(from_row, str):
            from_row = find_indexes(self.row_names, from_row)[0]
        if isinstance(to_row, str):
            to_row = find_indexes(self.row_names, to_row)[0]
        if isinstance(from_column, str):
            from_column = find_indexes(self.column_names, from_column)[0]
        if isinstance(to_column, str):
            to_column = find_indexes(self.column_names, to_column)[0]

        if to_row < 0:
            to_row += self.row_count + 1
        if to_column < 0:
            to_column += self.column_count + 1

        result = self.iloc[from_row:to_row, from_column:to_column]

        return self._df_query_return(result, pandas_return_type)

    def where(self, column_name: str, some_value: object, pandas_return_type: bool = True) -> pd.DataFrame:
        if isinstance(some_value, str):
            result = self[self[column_name].str.contains(some_value, na=False)]
        else:
            result = self.loc[self[column_name] == some_value]

        return self._df_query_return(result, pandas_return_type)

    def where_not(self, column_name: str, some_value: object, pandas_return_type: bool = True) -> pd.DataFrame:
        if isinstance(some_value, str):
            result = self[~self[column_name].isin([some_value])]
        else:
            result = self.loc[self[column_name] != some_value]

        return self._df_query_return(result, pandas_return_type)

    def random_row(self) -> pd.DataFrame:
        return self.iloc[np_random.randint(0, self.shape[0] - 1)]

    def regress(self, y: str or int, x: list or str or int, logit: bool = False) -> LinearModel or LogisticModel:
        if logit:
            return LogisticModel(dependent=self.select(column=y), independent=self.select(column=x))
        else:
            return LinearModel(dependent=self.select(column=y), independent=self.select(column=x))

    # Plotting
    def line_plot(self, *y: int or str, x: int or str or list = None) -> LinePlot:
        """
        Creates a Line Plot with y as response variable(s) and x as explanatory variable.
        :param y: Column name(s)/index(es) of response variable(s)
        :param x: Column name/index of explanatory variable
        :return: LinePlot with y as response variable(s) and x as explanatory variable.
        """
        return LinePlot(x=self.select(column=x) if x is not None else self.row_names,
                        y=[pd.Series(column[1]) for column in self.select(column=list(y)).iteritems()],
                        y_label=y[0] if len(y) == 1 and isinstance(y[0], str) else None,
                        x_label=x if isinstance(x, str) else None)

    def bar_plot(self, *bars: int or str, x: int or str or list = None) -> BarPlot:
        """
        Creates a Bar Plot with y as response variable(s) and x as explanatory variable.
        :param bars: Column name(s)/index(es) of response variable(s)
        :param x: Column name/index of explanatory variable
        :return: BarPlot with y as response variable(s) and x as explanatory variable.
        """
        return BarPlot(x=self.select(column=x) if x is not None else self.row_names if len(bars) > 0 else self.column_names,
                       y=[pd.Series(column[1]) for column in self.select(column=list(bars)).iteritems()] if len(bars) > 0 else [self.sum()],
                       y_label=bars[0] if len(bars) == 1 and isinstance(bars[0], str) else None,
                       x_label=x if isinstance(x, str) else None)

    def pie_plot(self, *slices) -> PiePlot:
        """
        Creates a Pie Plot for the slice(s) specified.
        :param slices: Column name(s)/index(es) to use for PiePlot sections
        :return: PiePlot for the slice(s) specified.
        """
        if len(slices) > 1:
            return PiePlot(self.select(column=list(slices)).sum().to_dict())
        elif len(slices) == 0:
            return PiePlot(self.sum().to_dict())
        else:
            return PiePlot(self.select(column=slices[0]).value_counts().to_dict())

    def scatter_plot(self, *y: int or str, x: int or str or list = None, best_fit_line: bool = False) -> ScatterPlot:
        """
        Creates a Scatter Plot with y as response variable(s) and x as explanatory variable.
        :param y: Column name(s)/index(es) of response variable(s)
        :param x: Column name/index of explanatory variable
        :param best_fit_line: flag indicating whether or not to include a best fit line in the ScatterPlot
        :return: ScatterPlot with y as response variable(s) and x as explanatory variable.
        """
        return ScatterPlot(x=self.select(column=x) if x is not None else self.row_names,
                           y=[pd.Series(column[1]) for column in self.select(column=list(y)).iteritems()],
                           y_label=y[0] if len(y) == 1 and isinstance(y[0], str) else None,
                           x_label=x if isinstance(x, str) else None,
                           best_fit_line=best_fit_line)

    def histogram(self, y: int or str, bins: int = 10) -> Histogram:
        return Histogram(y=self.select(column=y), bins=bins)

    def where_null(self, column: str, pandas_return_type: bool = True) -> DF or pd.DataFrame:
        result = self[self[column].isnull()]
        return self._df_query_return(result, pandas_return_type)

    def where_not_null(self, column: str, pandas_return_type: bool = True) -> DF or pd.DataFrame:
        result = self[self[column].notna()]
        return self._df_query_return(result, pandas_return_type)

    def group_by(self, column: int or str = None, row: int or str = None, date_grouping: str = None) -> dict:
        """
        Returns a dictionary object that groups on a Row/Column, using the grouping values as keys, pointing to Table objects containing the Row(s)/Column(s) that contain the key value.
        :param column: Column to group on
        :param row: Row to group on
        :param date_grouping: type of date grouping (e.g. "day", "month", "year")
        :return: Dictionary containing values grouped by (keys) and items belonging to that grouping (values).
        """
        if date_grouping is None:
            return {name: self.select(column=indexes, pandas_return_type=False)
                    if row is not None else self.select(row=indexes, pandas_return_type=False)
                    for name, indexes in index_dictionary(
                    (self.select(row=row, pandas_return_type=True) if row is not None
                     else self.select(column=column, pandas_return_type=True)).values).items()}
        else:
            grouping = DATE_GROUPINGS.get(date_grouping)
            if grouping is None:
                raise Exception(f"Invalid date grouping type \"{date_grouping}\"")
            if column is None:
                raise Exception("Cannot group on a Row of dates")
            date_group_by_object = self.groupby(pd.to_datetime(self.select(column=column)).dt.strftime(grouping))
            return {date_key: DF(date_group_by_object.get_group(date_key)) for date_key in list(date_group_by_object.groups.keys())}

    def sum_product(self, *columns: int or str) -> int or float:
        product_column = pd.Series([1]*self.row_count)
        for column in columns:
            product_column *= self.select(column=column, pandas_return_type=True)
        return product_column.sum()

    @reversible
    def update(self, row: int or str = None, column: int or str = None, to: object = None, in_place: bool = True) -> DF | None:
        if in_place:
            if column is not None:
                if row is not None:
                    if isinstance(column, str):
                        if isinstance(row, str):
                            self.loc[row, column] = to
                        else:
                            self.iloc[row, find_indexes(self.column_names, column)] = to
                    else:
                        if isinstance(row, str):
                            self.iloc[find_indexes(self.row_names, row), column] = to
                        else:
                            self.iloc[row, column] = to
                else:
                    if isinstance(column, str):
                        self.loc[self.row_names, column] = to
                    else:
                        self.iloc[create_index_list(self.row_count), column] = to
            elif row is not None:  # todo
                if isinstance(row, str):
                    pass
                else:
                    pass
            else:
                pass  # Note: if row is None and column is None, Exception is raised in @reversible. No need to raise one here.
        else:
            _df = DF(self)
            _df.update(row=row, column=column, to=to, in_place=True)
            return _df

    @reversible
    def append(self, row: pd.Series = None, column: pd.Series = None, in_place: bool = True) -> DF | None:
        if row is not None and column is None:
            if in_place:
                self._append_row(row, in_place)
            else:
                return self._append_row(row, in_place)
        elif row is None and column is not None:
            if column.name is None:  # without a column name, undo() does not work because it calls self.drop(None)
                column.name = self._create_column_name()
            if in_place:
                self._append_column(column, in_place)
            else:
                return self._append_column(column, in_place)
        elif row is not None and column is not None:
            if len(row) > len(column):
                if in_place:
                    self._append_column(column, in_place)
                    self._append_row(row, in_place)
                else:
                    return DF(copy(self))._append_column(column, in_place)._append_row(row, in_place)
            else:
                if in_place:
                    self._append_row(row, in_place)
                    self._append_column(column, in_place)
                else:
                    return DF(copy(self))._append_row(row, in_place)._append_column(column, in_place)
        else:
            self._remove_latest_unwind_operation()
            raise Exception("row or column parameter must be set")

    @reversible
    def insert(self, index: int or list, row: pd.Series or pd.DataFrame = None, column: pd.Series or pd.DataFrame = None, in_place: bool = True) -> DF | None:
        def insert_single_row_object(single_index: int, single_row: pd.Series or pd.DataFrame) -> None:
            before = self.slice(to_row=single_index, pandas_return_type=True)
            after = self.slice(from_row=single_index, pandas_return_type=True)
            if isinstance(single_row, pd.Series):
                if single_row.name is None or (isinstance(single_row.name, int) and single_row.name < self.row_count):
                    single_row.name = self.row_count
                self.__init__(overwrite_data=pd.concat([before, pd.DataFrame(single_row).transpose(), after], axis=0))
            else:  # pd.Dataframe
                self.__init__(overwrite_data=pd.concat([before, self._update_row_names(single_row), after], axis=0))

        def insert_discontinuous_row_objects(index_list: list, df: DF) -> None:
            if len(index_list) != df.row_count:
                raise Exception("To insert discontinuous row objects, row count must match the number of index positions passed")
            for counter, insert_index in enumerate(index_list):
                insert_single_row_object(insert_index, df.select(row=counter, pandas_return_type=True))

        def insert_single_column_object(single_index: int, single_column: pd.Series or pd.DataFrame) -> None:
            before = self.slice(to_column=single_index, pandas_return_type=True)
            after = self.slice(from_column=single_index, pandas_return_type=True)
            if isinstance(single_column, pd.Series):
                if single_column.name is None:
                    single_column.name = self.column_count
                self.__init__(overwrite_data=pd.concat([before, single_column, after], axis=1))
            else:  # pd.DataFrame
                self.__init__(overwrite_data=pd.concat([before, self._update_column_names(single_column), after], axis=1))

        def insert_discontinuous_column_objects(index_list: list, df: DF) -> None:
            if len(index_list) != df.column_count:
                raise Exception("To insert discontinuous column objects, column count must match the number of index positions passed")
            for counter, insert_index in enumerate(index_list):
                insert_single_column_object(insert_index, df.select(column=counter, pandas_return_type=True))

        if in_place:
            if row is not None:
                if isinstance(index, int):
                    insert_single_row_object(index, row)
                elif is_continuous_list(index):
                    insert_single_row_object(index[0], row)
                else:
                    insert_discontinuous_row_objects(index, DF(row))
            elif column is not None:
                if isinstance(index, int):
                    insert_single_column_object(index, column)
                elif is_continuous_list(index):
                    insert_single_column_object(index[0], column)
                else:
                    insert_discontinuous_column_objects(index, DF(column))
            else:
                self._remove_latest_unwind_operation()
                raise Exception("Must pass row OR column")
        else:
            _df = DF(self)
            _df.insert(index=index, row=row, column=column, in_place=True)
            return _df

    @reversible
    def remove(self, row: int or str or list = None, column: int or str or list = None, in_place: bool = True) -> DF | None:
        if row is not None and column is not None:  # todo: remove this once .insert() method allows for row and column to be passed
            raise Exception("Please supply either a row or column argument, but not both")
        if in_place:
            if row is not None:
                if isinstance(row, int):  # drop by row index
                    rows_to_drop = self.row_names[row]
                elif isinstance(row, list):
                    if any([isinstance(row_selector, str) for row_selector in row]):
                        rows_to_drop = row  # drop by row names
                    else:
                        rows_to_drop = [self.row_names[r] for r in row]  # drop by row indexes
                else:  # drop by row name
                    rows_to_drop = row
                self.drop(rows_to_drop, axis=0, inplace=True)
            elif column is not None:
                if isinstance(column, int):  # drop by column index
                    columns_to_drop = self.column_names[column]
                elif isinstance(column, list):
                    if any([isinstance(column_selector, str) for column_selector in column]):
                        columns_to_drop = column  # drop by column names
                    else:
                        columns_to_drop = [self.column_names[r] for r in column]  # drop by column indexes
                else:  # drop by column name
                    columns_to_drop = column
                self.drop(columns_to_drop, axis=1, inplace=True)
            else:
                pass  # Note: if row is None and column is None, Exception is raised in @reversible. No need to raise one here.
        else:
            _df = DF(self.dataframe)
            _df.remove(row=row, column=column, in_place=True)
            return _df

    def undo(self) -> None:
        """ Reverts the most recent change to self """
        try:
            unwind_object: Unwind = self._unwind.pop()
        except IndexError:
            raise Exception("There are no DF manipulations to undo")
        self.__getattribute__(unwind_object.function)(**unwind_object.args[0])

    @reversible
    def reset(self, in_place: bool = True) -> DF | None:
        if in_place:
            self._set_underlying_dataframe(data=self.data)
        else:
            return DF(data=self.data)

    def _set_underlying_dataframe(self, data) -> None:
        super().__init__(data=data)

    def _append_row(self, row: pd.Series, in_place: bool) -> DF | None:
        if in_place:
            self._set_underlying_dataframe(data=super().append(other=row, ignore_index=True))
        else:
            _df = copy(self)
            _df = _df.append(other=row, ignore_index=True)
            return DF(_df)

    def _append_column(self, column: pd.Series, in_place: bool) -> DF | None:
        if in_place:
            super().insert(loc=self.column_count, column=column.name, value=column, allow_duplicates=True)
        else:
            _df = copy(self)
            _df.insert(loc=self.column_count, column=column.name, value=column, allow_duplicates=True)
            return DF(_df)

    def _create_column_name(self) -> str:
        new_cols = len([col for col in self.column_names if "new_col" in col])
        return f"new_col{new_cols + 1}"

    def _remove_latest_unwind_operation(self):
        del self._unwind[-1]

    def _update_row_names(self, rows: pd.DataFrame) -> pd.DataFrame:
        offset, name_changes = 0, {}
        for row in rows.iterrows():
            if isinstance(row[1].name, int) and row[1].name == offset:  # update row names if it appears they have generic index (i.e., [0, 1, 2...])
                name_changes[row[1].name] = self.row_count + offset
                offset += 1
        return rows.rename(name_changes, inplace=False, axis=0)

    def _update_column_names(self, columns: pd.DataFrame) -> pd.DataFrame:
        offset, name_changes = 0, {}
        for column in columns:
            if isinstance(columns[column].name, int) and columns[column].name == offset:  # update column names if it appears they have generic index (i.e., [0, 1, 2...])
                name_changes[columns[column].name] = self.column_count + offset
                offset += 1
        return columns.rename(name_changes, inplace=False, axis=1)

    @staticmethod
    def _df_query_return(result, pandas_return_type: bool):
        if pandas_return_type or not isinstance(result, (pd.Series, pd.DataFrame)):
            return result
        else:
            return DF(result)

    @staticmethod
    def _names_supplied(selector: int or str or list) -> bool:
        if isinstance(selector, list):
            return isinstance(selector[0], str)
        else:
            return isinstance(selector, str)

    @staticmethod
    def _create_str_column_names(columns: int) -> list:
        letters, headers = [letter for letter in ALPHABET], []
        for column_index in range(columns):
            cycles = column_index // len(letters)
            if cycles == 0:
                headers.append(letters[column_index])
            elif cycles <= len(letters):
                headers.append(letters[cycles - 1] + letters[column_index % len(letters)])
            else:
                headers.append(letters[(cycles // len(letters)) - 1] + letters[(cycles % len(letters)) - 1] + letters[
                    column_index % len(letters)])
        return headers

    @property
    def is_empty(self) -> bool:
        return self.empty

    @property
    def rows(self) -> list:
        return [pd.Series(row[1]) for row in self.iterrows()]

    @property
    def row_names(self):
        return [row.name for row in self.rows]

    @property
    def row_count(self) -> int:
        return len(self)

    @property
    def column_names(self) -> list:
        return list(self.columns)

    @property
    def column_count(self) -> int:
        return len(self.columns)


class RandomDF(DF):
    def __init__(self,
                 rows: int = None,
                 columns: int = None,
                 data_type: type = float,
                 distribution: str = 'uniform'):
        """
        todo: add summary of class
        :param rows: Number of rows to create RandomDF with. If None, a random number of rows between 5 and 200 will be chosen
        :param columns: Number of columns to create RandomDF with. If None, a random number of columns between 5 and 200 will be chosen
        :param data_type: Type of random object to create and populate DF with. Options are float (default), int, and str
        :param distribution: Type of distribution to draw random numbers from (ignored if data_type=str. Options are uniform (default) and normal
        """
        if rows is None:
            rows = np_random.randint(5, 200)
        if columns is None:
            columns = np_random.randint(5, 200)
        self.number_of_rows = rows
        self.number_of_columns = columns
        self.data_type = data_type
        self.distribution = distribution
        column_names = self._create_str_column_names(columns)
        data = {}
        for column_count in range(self.number_of_columns):
            data[column_names[column_count]] = []
            for _ in range(self.number_of_rows):
                data[column_names[column_count]].append(self._get_random_data_point(data_type, distribution))
        self._set_underlying_dataframe(data=data)

    def regenerate(self,
                   number_of_rows: int = None,
                   number_of_columns: int = None,
                   data_type: type = None,
                   distribution: str = None) -> None:
        self._set_underlying_dataframe(data=RandomDF(rows=number_of_rows if number_of_rows is not None else self.number_of_rows,
                                       columns=number_of_columns if number_of_columns is not None else self.number_of_columns,
                                       data_type=data_type if data_type is not None else self.data_type,
                                       distribution=distribution if distribution is not None else self.distribution))

    # Static methods
    @staticmethod
    def _get_random_data_point(data_type: type, distribution: str) -> object:
        if data_type is str:
            return np_random.choice([letter for letter in ALPHABET])
        elif data_type in {float, int}:
            if distribution not in DISTRIBUTIONS:
                raise ValueError(f"Cannot draw random number from {distribution} distribution. "
                                 f"Available distributions include {DISTRIBUTIONS}")
            return data_type(np_random.__getattribute__(distribution)())
        elif data_type is None:
            return None
        else:
            raise TypeError(f"Invalid type for RandomDF values {data_type}")


class EmptyDF(DF):
    def __init__(self,
                 number_of_rows: int = None,
                 number_of_columns: int = None,
                 column_headers: bool = True):
        self.number_of_rows = number_of_rows
        self.number_of_columns = number_of_columns
        data = {}
        if number_of_columns is not None:
            if column_headers:
                column_names = self._create_str_column_names(number_of_columns)
                for column_count in range(self.number_of_columns):
                    data[column_names[column_count]] = [None for _ in range(self.number_of_rows)]
            else:
                data = [[None for _ in range(number_of_columns)] for _ in range(number_of_rows)]
        self._set_underlying_dataframe(data=data)
